from rest_framework_simplejwt.serializers import TokenObtainPairSerializer
from rest_framework import serializers
from django.contrib.auth.models import User
from clusterCore.exceptions import *
from datetime import datetime
from django.utils import timezone
from  django.db.models import Case, When
from django.db import connection

from clusterCore.enums import DBType, ClusterRoles, Severity, Status, Node_Size_Allowed
from clusterCore.models import (OracleSystemConfig, Profile,  CustomerKey, SystemConfig, TaskLog, c123_cluster, c123_database, c123_inventory, c123_node,
                                c123_host_monitor, c123_dashboard_item,
                                c123_host,TaskStepLog, ClusterSizing, OracleClusterSizing,
                                Task,  UserPrivileges, UserNotifications, IntegratedService)
from django.db.models.expressions import RawSQL
import json
from django.contrib.auth.models import update_last_login
PRIMARY_CREATION_FAILED = "Primary {} are failed to create.Could not add secondary nodes."
PRIMARY_CREATING = "Primary {} are now creating.Could not add secondary nodes."
PRIMARY_GONE = "Primary {} is not available."
FORBID_APP_HOST = "{} is app host.Could not register nodes here."
RECREATION_CONFLICT = "{} node already exists at {}."
OS_CONFLICT = 'Primary ({} {}) and the new node({} {}) have not the same os version'
OS_UNSUPPORTED = "{} {} does not support {} {}"

def get_create_node_response(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        task = serializer.save()
        from clusterCore.tasks import append_status_cmd
        return append_status_cmd(request, task)

def format_size(size):
    #format size to display in user interface
    units = ['KB', 'MB', 'GB', 'TB']  # List of units
    if size == '':
        size = 0.0
    size = float(size)
    # Loop through the units and divide the size by 1024 until it's less than 1024
    for unit in units:
        if size < 1024.0:
            return f'{size:.2f} {unit}'
        size /= 1024.0
    return f'{size:.2f} {units[-1]}'  # Return the size in the largest unit




def get_size_scheme_util(obj):
    if not obj:
        return None
    
    match obj.type:
        case DBType.ORACLE_CDB.value | DBType.ORACLE_PDB.value:
            cs, created = OracleClusterSizing.objects.get_or_create()
        case DBType.POSTGRES_K8S.value:
             cs, created = ClusterSizing.objects.get_or_create(db_type = DBType.POSTGRES.value)
        case DBType.MONGO_K8S.value:
             cs, created = ClusterSizing.objects.get_or_create(db_type = DBType.MONGO.value)
        case DBType.HAPROXY.value:
            return None
        case _:
             cs, created = ClusterSizing.objects.get_or_create(db_type = obj.type)


    avail_schemes = {
        Node_Size_Allowed.SMALL.name: {
            'cpu': cs.small_cpu,
            'disk': cs.small_disk,
            'memory': cs.small_memory,
            'backup': cs.small_backup_keep_days,
            'export': cs.small_export_keep_days,
        },
        Node_Size_Allowed.MEDIUM.name: {
            'cpu': cs.medium_cpu,
            'disk': cs.medium_disk,
            'memory': cs.medium_memory,
            'backup': cs.medium_backup_keep_days,
            'export': cs.medium_export_keep_days,
        },
        Node_Size_Allowed.LARGE.name: {
            'cpu': cs.large_cpu,
            'disk': cs.large_disk,
            'memory': cs.large_memory,
            'backup': cs.large_backup_keep_days,
            'export': cs.large_export_keep_days,
        },
        Node_Size_Allowed.VERY_LARGE.name: {
            'cpu': cs.very_large_cpu,
            'disk': cs.very_large_disk,
            'memory': cs.very_large_memory,
            'backup': cs.very_large_backup_keep_days,
            'export': cs.very_large_export_keep_days,
        }
    }
    if obj.ex_params and obj.ex_params.get('sizing'):
        sizing =  obj.ex_params.get('sizing')
    else:
        sizing = obj.sizing

    return avail_schemes.get(sizing)

class c123_AddIntegratedServiceSerializer(serializers.ModelSerializer):
    class Meta:
        model = IntegratedService
        fields = '__all__'
    def create(self, validated_data):

        integrated_service_exists = IntegratedService.objects.filter(name = validated_data['name'], host_name = validated_data['host_name']).exists()
        if integrated_service_exists:
            raise APIException406("Integrated service with such name and hostname has been already registered in system")
        return IntegratedService.objects.create(**validated_data)

class SelfTokenObtainPairSerializer(TokenObtainPairSerializer):
    @classmethod
    def get_token(cls, user):
        token = super().get_token(user)
        token['username'] = user.username
        update_last_login(None, user)
        return token
class c123_HostSerializer(serializers.ModelSerializer):
    class Meta:
        model= c123_host
        fields ='__all__'

class c123_IntegratedServiceSerializer(serializers.ModelSerializer):
    class Meta:
        model= IntegratedService
        fields ='__all__'

class MongoShardedPrimarySerializer(serializers.ModelSerializer):
    host = serializers.SerializerMethodField()
    nodes = serializers.SerializerMethodField()

    def get_host(self, obj):
        return obj.host.name

    def get_nodes(self, obj):
        return c123_NodeSerializer(c123_node.objects.filter(cluster=obj.cluster,cluster_role=ClusterRoles.STANDBY.value, ex_params__replicaset_name=obj.ex_params.get('replicaset_name'), ind_deleted = False), many=True).data

    class Meta:
        model = c123_node
        fields ='__all__'

class c123_dashboardSerializer(serializers.ModelSerializer):
    org_id = serializers.SerializerMethodField()
    node_name = serializers.SerializerMethodField()
    class Meta:
        model = c123_dashboard_item
        fields = '__all__'

    def get_org_id(self, obj):
        return obj.created_by.profile.org_id.org_id

    def get_node_name(self, obj):
        return obj.node_id.name

class TaskSerializer(serializers.ModelSerializer):
    severity = serializers.SerializerMethodField()
    class Meta:
        model = Task
        fields = '__all__'

    def get_severity(self, obj):
        latest_log = TaskLog.objects.filter(task_id = obj.pk).order_by("created_date").last()
        if latest_log:
            return latest_log.status
        return ""
    
    
class c123_ClusterStatusSerializer(serializers.ModelSerializer):
    severity = serializers.SerializerMethodField()
    status = serializers.SerializerMethodField()
    nodes = serializers.SerializerMethodField()
    what = serializers.CharField(source='get_type_display', read_only=True)

    def get_nodes(self, obj):
        # Fetching nodes related to the cluster
        nodes = c123_node.objects.filter(cluster=obj, ind_deleted=False)

        # Prepare the raw SQL query to get the last alert for each node
        query = """
            SELECT node_id, message
            FROM app_123monitor.rp_alert
            WHERE node_id IN %s
            ORDER BY alert_id DESC
        """
        
        # Get a list of node ids for the query
        node_ids = tuple(nodes.values_list('id', flat=True))  # Create a tuple
        
        # Fetch last alerts from the rp_alert table
        with connection.cursor() as cursor:
            cursor.execute(query, [node_ids])  # Pass the tuple directly
            alert_rows = cursor.fetchall()

        # Create a dictionary to map node_id to the last alert message
        node_alert_map = {node_id: alert_message for node_id, alert_message in alert_rows}

        # Annotate nodes with the last alert message from rp_alert
        for node in nodes:
            node.alert_message = node_alert_map.get(node.id, None)

        # Serialize nodes with the last alert message included
        node_data = c123_NodeSerializer(nodes, many=True).data
        return node_data

    def get_status(self, obj):
        nodes = c123_node.objects.filter(cluster=obj, ind_deleted=False)
        if not nodes.filter(severity=Severity.OK.value).exists():
            return Status.DOWN.value
        if nodes.filter(severity=Severity.WARNING.value).exists():
            return Status.WARNING.value
        return Status.UP.value

    def get_severity(self, obj):
        nodes = c123_node.objects.filter(cluster=obj, ind_deleted=False)

        if not nodes.filter(cluster_role=ClusterRoles.PRIMARY.value, severity__in=[Severity.OK.value, Severity.WARNING.value]).exists():
            return Severity.NOT_OK.value

        if nodes.filter(severity__in=[Severity.NOT_OK.value, Severity.WARNING.value]).exists():
            return Severity.WARNING.value

        return Severity.OK.value

    class Meta:
        model = c123_cluster
        fields = ["id", "nodes", "what", "severity", "name", "type", "status"]

class c123_ClusterSerializer(serializers.ModelSerializer):
    size_scheme = serializers.SerializerMethodField()
    class Meta:
        model = c123_cluster
        fields = '__all__'

    def get_size_scheme(self, obj):
        if not obj.ex_params or not obj.ex_params.get("sizing"):
            return None
        match obj.type:
            case DBType.REDIS.value | DBType.MONGO.value:
                    return get_size_scheme_util(obj)
            case _ :
                return None



class c123_DatabaseSerializer(serializers.ModelSerializer):
    host = serializers.SerializerMethodField()
    display_size = serializers.SerializerMethodField()
    alert = serializers.SerializerMethodField()
    cluster = serializers.SerializerMethodField()
    size_scheme = serializers.SerializerMethodField()
    
    class Meta:
        model= c123_database
        fields ='__all__'
    
    def get_alert(self, obj):
        return 'No alerts'
    
    def get_cluster(self, obj):
        if obj.node:
            return obj.node.cluster.id
        return None
    
    def get_host(self, obj):
        if obj.node:
            return obj.node.name
        return None
    
    def get_display_size(self, obj):
        return '0 KB' if not obj.size else format_size(obj.size)
    
    def get_size_scheme(self, obj):
        return get_size_scheme_util(obj.node)
    
class c123_AllNodeSerializer(serializers.ModelSerializer):
    host = c123_HostSerializer()
    cluster = c123_ClusterSerializer()
    class Meta:
        model = c123_node
        fields = ['name', 'host', 'type', 'cluster', 'created_by', 'last_update_by', 'cluster_role', 'database_user', 'database_pwd', 'port']
    def create(self, validated_data):
        cluster_role = validated_data.pop('cluster_role')
        if cluster_role == ClusterRoles.CDB:
            cluster_data = validated_data.pop('cluster')
            cluster,created = c123_cluster.objects.get_or_create(**cluster_data)
            host_data = validated_data.pop('host')
            host,created = c123_host.objects.get_or_create(**host_data)
            return c123_node.objects.create(host=host, cluster=cluster, **validated_data)
        else:
           """  host_data = validated_data.pop('host')
            host,created = c123_host.objects.get_or_create(**host_data)
            return c123_node.objects.create(host=host,**validated_data) """
# class c123_PDBSerializer(serializers.ModelSerializer):
#     class Meta:
#         model = c123_node
#         fields = ['name', 'host', 'type', 'cluster', 'created_by', 'last_update_by', 'database_user', 'database_pwd', 'port', 'ex_params']
#     def create(self, validated_data):
#         return c123_node.objects.create(cluster_role="PDB", **validated_data)
class CustomerKeySerializer(serializers.ModelSerializer):
    class Meta:
        model = CustomerKey
        fields = '__all__'


class c123_NodeSerializer(serializers.ModelSerializer):
    what = serializers.CharField(source='get_type_display', read_only=True)
    location = serializers.SerializerMethodField()
    host = serializers.CharField(source='host.name', default='', read_only=True)
    created_by = serializers.CharField(source='created_by.username', default='', read_only=True)
    host_id = serializers.IntegerField(read_only=True)
    cdb_name = serializers.CharField(read_only=True)
    size_scheme = serializers.SerializerMethodField()
    display_size = serializers.CharField(read_only=True)
    slaves_count = serializers.IntegerField(read_only=True)
    primary_count = serializers.IntegerField(read_only=True)
    has_trial = serializers.BooleanField(read_only=True)
    cluster_content = c123_ClusterSerializer(source='cluster', read_only=True)
    alert = serializers.SerializerMethodField()

    def get_alert(self, obj):
        try:
            cursor = connection.cursor()
        except Exception as e:
            error = e.strerror if hasattr(e, "strerror") else e.args[0]
            print(error)
            return "No alerts"
        try:
            cursor.execute(f"SELECT message FROM app_123monitor.rp_alert WHERE node_id = {obj.id} ORDER BY alert_id DESC;")
        except Exception as e:
            error = e.strerror if hasattr(e, "strerror") else e.args[0]
            print(error)
            return "No alerts"
        try:
            alert = cursor.fetchone()
        except Exception as e:
            error = e.strerror if hasattr(e, "strerror") else e.args[0]
            print(error)
            return "No alerts"
        
        if not alert:
            return "No alerts"
        return alert[0]

    def get_location(self, obj):
        templates = {
            DBType.ORACLE_CDB.value: "{name}@{host}({role})",
            DBType.ORACLE_PDB.value: "{name}@{host}@{port}({role})",
        }
        return templates.get(obj.type, "{name}({role})").format(
            name=obj.name, host=obj.host.name if obj.host else "", port=obj.port, role=obj.cluster_role
        )

    def get_size_scheme(self, obj):
        return get_size_scheme_util(obj)

    class Meta:
        model = c123_node
        fields = '__all__'

class c123_CreateHostSerializer(serializers.ModelSerializer):
    class Meta:
        model=c123_host
        fields = ['name', 'ssh_port', 'username', 'password', 'private_key', 'ex_params']
class c123_CreateTempClusterSerializer(serializers.ModelSerializer):
        class Meta:
            model= c123_cluster
            fields = ['name', 'type', 'ex_params']
class c123_HostMonitorSerializer(serializers.ModelSerializer):
    class Meta:
        model = c123_host_monitor
        fields = '__all__'
class c123_TaskLogSerializer(serializers.ModelSerializer):
    what = serializers.SerializerMethodField()
    when = serializers.SerializerMethodField()
    created_by = serializers.SerializerMethodField()
    def get_created_by(self,obj):
        if(obj.created_by):
            return obj.created_by.username
        return ""
    def get_what(self,obj):
        task = Task.objects.filter(pk = obj.task_id).first()
        if task:
            return task.monitor_name if task.monitor_name != None else task.displayment
        return ""

    def get_when(self,obj):
        when =obj.created_date
        return when

    class Meta:
        model = TaskLog
        fields = '__all__'
class c123_TaskLogTextSerializer(serializers.ModelSerializer):
    single_field = serializers.SerializerMethodField()
    def get_single_field(self,obj):
        single_field = str(obj.created_date) + "    " + str(obj.task)
        return single_field

    class Meta:
        model = TaskLog
        fields = ['single_field']
class c123_TaskStepLogSerializer(serializers.ModelSerializer):
    what = serializers.SerializerMethodField()
    when = serializers.SerializerMethodField()
    def get_what(self,obj):
        what = obj.task
        return what

    def get_when(self,obj):
        when =obj.created_date
        return when

    class Meta:
        model = TaskStepLog
        fields = ['id', 'task_id', 'task_log_id', 'status', 'when', 'what', 'created_date', 'duration']
class c123_TaskStepLogDetailSerializer(serializers.ModelSerializer):
        created_by = serializers.SerializerMethodField()

        def get_created_by(self, obj):
            created_by = ""
            task_log = TaskLog.objects.filter(task_log_id = obj.task_log_id).first()
            if task_log:
                if task_log.created_by:
                    created_by = str(task_log.created_by.username)

            return created_by

        class Meta:
            model = TaskStepLog
            fields = ['status', 'created_date', 'task', 'step', 'retcode', 'retmsg', 'duration', 'retoutput', 'created_by']
class ClusterSizingSerializer(serializers.ModelSerializer):

    class Meta:
        model = ClusterSizing
        fields = '__all__'
class OracleClusterSizingSerializer(serializers.ModelSerializer):

    class Meta:
        model = OracleClusterSizing
        fields = '__all__'
class UserPrivilegesSerializer(serializers.ModelSerializer):

      class Meta:
        model = UserPrivileges
        fields = '__all__'
class UserNotificationsSerializer(serializers.ModelSerializer):

      class Meta:
        model = UserNotifications
        fields = '__all__'
class c123_UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ('username', 'email', 'id')
class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ('username', 'date_joined', 'last_login', 'id')
class c123_inventory_serializer(serializers.ModelSerializer):
    created_by = serializers.SerializerMethodField()
    node_name = serializers.SerializerMethodField()
    host = serializers.SerializerMethodField()
    host_name = serializers.SerializerMethodField()
    display_size = serializers.SerializerMethodField()
    cluster = serializers.SerializerMethodField()
    run_log = serializers.SerializerMethodField()

    def get_run_log(self, obj):
        if obj.run_log:
            return obj.run_log
        if not obj.ex_params or not obj.ex_params.get('task_id'):
            return ''
        last_step_log = TaskStepLog.objects.filter(task_id = obj.ex_params.get('task_id')).last()
        if last_step_log:
            return last_step_log.retmsg
        return ''
        

    def get_display_size(self, obj):
        return '0 KB' if not obj.ex_params or not obj.ex_params.get('size') else format_size(obj.ex_params['size'])

    def get_created_by(self, obj):
        user = User.objects.filter(pk = obj.created_by.pk).first()
        return user.username if user else ""

    def get_cluster(self, obj):
      if obj.node and obj.node.cluster:
          return obj.node.cluster.id
      elif obj.database and obj.database.node and obj.database.node.cluster:
          return obj.database.node.cluster.id
      return None

    def get_node_name(self, obj):
        if obj.database:
             return obj.database.name
        elif obj.node:
            return obj.node.name
    def get_host_name(self, obj):
        if obj.database:
             return obj.database.node.host.name
        elif obj.node:
            return obj.node.host.name
    def get_host(self, obj):
        if obj.node:
            node = obj.node
            return node.name + "@" + str(node.host) if "ORACLE" in node.type else f'{node.host.name}:{node.port}'
        elif obj.database:
            node = obj.database.node
            return f'{node.host.name}:{node.port}'

    def create(self, validated_data):
        return c123_inventory(**self.validated_data)
    class Meta:
        model = c123_inventory
        fields =  '__all__'
class CurrentUserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ('username',)
class SystemConfigSerializer(serializers.ModelSerializer):

     class Meta:
        model = SystemConfig
        fields =  '__all__'
class OracleSystemConfigSerializer(serializers.ModelSerializer):
     class Meta:
        model = OracleSystemConfig
        fields =  '__all__'
class ConfigSerializer(serializers.Serializer):
    current_user = CurrentUserSerializer()
    mariadb_cluster_sizing = ClusterSizingSerializer()
    mongo_cluster_sizing = ClusterSizingSerializer()
    mysql_cluster_sizing = ClusterSizingSerializer()
    elasticsearch_cluster_sizing = ClusterSizingSerializer()
    postgres_cluster_sizing = ClusterSizingSerializer()
    redis_cluster_sizing = ClusterSizingSerializer()
    mssql_cluster_sizing = ClusterSizingSerializer()
    oracle_cluster_sizing = OracleClusterSizingSerializer()
    pg_sys = SystemConfigSerializer()
    mariadb_sys = SystemConfigSerializer()
    mysql_sys = SystemConfigSerializer()
    elasticsearch_sys = SystemConfigSerializer()
    mongo_sys = SystemConfigSerializer()
    redis_sys = SystemConfigSerializer()
    mssql_sys = SystemConfigSerializer()
    oracle_sys = OracleSystemConfigSerializer()
    licenses = CustomerKeySerializer(many = True)
    license_summary = CustomerKeySerializer()
    users = UserSerializer(many = True)
    def create(self, validated_data):
        current_user = CurrentUserSerializer(data=validated_data['current_user'])
        if current_user.is_valid():
            current_user.save()
        mongo_cluster_sizing = ClusterSizingSerializer(data=validated_data['mongo_cluster_sizing'])
        if mongo_cluster_sizing.is_valid():
            mongo_cluster_sizing.save()
        mysql_cluster_sizing = ClusterSizingSerializer(data=validated_data['mysql_cluster_sizing'])
        if mysql_cluster_sizing.is_valid():
            mysql_cluster_sizing.save()
        mariadb_cluster_sizing = ClusterSizingSerializer(data=validated_data['mariadb_cluster_sizing'])
        if mariadb_cluster_sizing.is_valid():
            mariadb_cluster_sizing.save()
        elasticsearch_cluster_sizing = ClusterSizingSerializer(data=validated_data['elasticsearch_cluster_sizing'])
        if elasticsearch_cluster_sizing.is_valid():
            elasticsearch_cluster_sizing.save()
        postgres_cluster_sizing = ClusterSizingSerializer(data=validated_data['postgres_cluster_sizing'])
        if postgres_cluster_sizing.is_valid():
            postgres_cluster_sizing.save()
        redis_cluster_sizing = ClusterSizingSerializer(data=validated_data['redis_cluster_sizing'])
        if redis_cluster_sizing.is_valid():
            redis_cluster_sizing.save()
        mssql_cluster_sizing = ClusterSizingSerializer(data=validated_data['mssql_cluster_sizing'])
        if mssql_cluster_sizing.is_valid():
            mssql_cluster_sizing.save()
        oracle_cluster_sizing= OracleClusterSizingSerializer(data=validated_data['oracle_cluster_sizing'])
        if oracle_cluster_sizing.is_valid():
            oracle_cluster_sizing.save()
        pg_sys = SystemConfigSerializer(data=validated_data['pg_sys'])
        if pg_sys.is_valid():
            pg_sys.save()
        mongo_sys = SystemConfigSerializer(data=validated_data['mongo_sys'])
        if mongo_sys.is_valid():
            mongo_sys.save()
        mariadb_sys = SystemConfigSerializer(data=validated_data['mariadb_sys'])
        if mariadb_sys.is_valid():
            mariadb_sys.save()
        mysql_sys = SystemConfigSerializer(data=validated_data['mysql_sys'])
        if mysql_sys.is_valid():
            mysql_sys.save()
        mssql_sys = SystemConfigSerializer(data=validated_data['mssql_sys'])
        if mssql_sys.is_valid():
            mssql_sys.save()
        elasticsearch_sys = SystemConfigSerializer(data=validated_data['elasticsearch_sys'])
        if elasticsearch_sys.is_valid():
            elasticsearch_sys.save()
        redis_sys = SystemConfigSerializer(data=validated_data['redis_sys'])
        if redis_sys.is_valid():
            redis_sys.save()
        oracle_sys = OracleSystemConfigSerializer(data=validated_data["oracle_sys"])
        if oracle_sys.is_valid():
            oracle_sys.save()
        users = UserSerializer(data=validated_data['users'], many=True)
        if users.is_valid():
            users.save()

        licenses = CustomerKeySerializer(validated_data['licenses'], many = True)
        license_summary = CustomerKeySerializer(validated_data['license_summary'])
        return self.__class__({'pg_sys': pg_sys, 'oracle_sys': oracle_sys, "redis_sys": redis_sys, 'mongo_sys': mongo_sys, "mariadb_sys": mariadb_sys, "mysql_sys": mysql_sys, "mssql_sys": mssql_sys, 'elasticsearch_sys': elasticsearch_sys,  'current_user': current_user, 'mongo_cluster_sizing': mongo_cluster_sizing, 'redis_cluster_sizing': redis_cluster_sizing,  'mariadb_cluster_sizing': mariadb_cluster_sizing, 'elasticsearch_cluster_sizing': elasticsearch_cluster_sizing, 'mysql_cluster_sizing': mysql_cluster_sizing, 'mssql_cluster_sizing': mssql_cluster_sizing, 'postgres_cluster_sizing': postgres_cluster_sizing, 'oracle_cluster_sizing': oracle_cluster_sizing, 'users': users, 'licenses': licenses, 'license_summary': license_summary})
def object_to_json(obj) -> str:
    """Serialize Django ORM objects to JSON"""
    obj_dict = obj.__dict__
    data = {}
    excluded_keys = {'_state', 'created_date', 'nodes'}
    for key, value in obj_dict.items():
        if key in excluded_keys:
            continue
        elif key == 'auto_delete' or key == 'time':
            value = str(value.replace(tzinfo=None)) if type(value) is datetime else None
        if key.endswith('_id_id'):
            key = key.replace('_id_id', '_id')
        data[key] = value
    return json.dumps(data, indent=4, sort_keys=True, default=str)
